@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2009-2014 Codership Oy <info@codership.com>
+ * Copyright (C) 2009-2019 Codership Oy <info@codership.com>
  */
 
 #include "gmcast.hpp"
@@ -46,6 +46,30 @@ static std::string get_scheme(bool use_ssl)
     return gu::scheme::tcp;
 }
 
+//
+// Check if the node should stay isolated.
+// Possible outcomes:
+// * Return false, node should continue reconnecting and accepting connections
+//   (isolate = 0)
+// * Return true, node should remain isolated (isolate = 1)
+// * Throw fatal exception to terminate the backend (isolate = 2)
+//
+static inline bool is_isolated(int isolate)
+{
+    switch (isolate)
+    {
+    case 1:
+        return true;
+    case 2:
+        gu_throw_fatal<< "Gcomm backend termination was "
+                      << "requested by setting gmcast.isolate=2.";
+        break;
+    default:
+        break;
+    }
+    return false;
+}
+
 gcomm::GMCast::GMCast(Protonet& net, const gu::URI& uri,
                       const UUID* my_uuid)
     :
@@ -77,7 +101,8 @@ gcomm::GMCast::GMCast(Protonet& net, const gu::URI& uri,
     remote_addrs_ (),
     addr_blacklist_(),
     relaying_     (false),
-    isolate_      (false),
+    isolate_      (0),
+    prim_view_reached_(false),
     proto_map_    (new ProtoMap()),
     relay_set_    (),
     segment_map_  (),
@@ -233,7 +258,7 @@ void gcomm::GMCast::set_initial_addr(const gu::URI& uri)
         {
             port = i->port();
         }
-        catch (gu::NotSet& )
+        catch (gu::NotSet&)
         {
             try
             {
@@ -243,6 +268,10 @@ void gcomm::GMCast::set_initial_addr(const gu::URI& uri)
             {
                 port = Defaults::GMCastTcpPort;
             }
+            catch (gu::NotSet&)
+            {
+                port = Defaults::GMCastTcpPort;
+            }
         }
 
         std::string initial_uri = uri_string(get_scheme(use_ssl_), host, port);
@@ -357,6 +386,91 @@ void gcomm::GMCast::close(bool force)
     proto_map_->clear();
     pending_addrs_.clear();
     remote_addrs_.clear();
+    prim_view_reached_ = false;
+}
+
+//
+// Private
+//
+
+
+// Find other local endpoint matching to proto
+static const Proto*
+find_other_local_endpoint(const gcomm::gmcast::ProtoMap& proto_map,
+                          const gcomm::gmcast::Proto* proto)
+{
+    for (gcomm::gmcast::ProtoMap::const_iterator i(proto_map.begin());
+         i != proto_map.end(); ++i)
+    {
+        if (i->second != proto &&
+            i->second->handshake_uuid() == proto->handshake_uuid())
+        {
+            return i->second;
+        }
+    }
+    return 0;
+}
+
+// Find other endpoint with same remote UUID
+static const Proto*
+find_other_endpoint_same_remote_uuid(const gcomm::gmcast::ProtoMap& proto_map,
+                                     const gcomm::gmcast::Proto* proto)
+{
+    for (gcomm::gmcast::ProtoMap::const_iterator i(proto_map.begin());
+         i != proto_map.end(); ++i)
+    {
+        if (i->second != proto &&
+            i->second->remote_uuid() == proto->remote_uuid())
+        {
+            return i->second;
+        }
+    }
+    return 0;
+}
+
+bool gcomm::GMCast::is_own(const gmcast::Proto* proto) const
+{
+    assert(proto->remote_uuid() != gcomm::UUID::nil());
+    if (proto->remote_uuid() != uuid())
+    {
+        return false;
+    }
+    return find_other_local_endpoint(*proto_map_, proto);
+}
+
+void gcomm::GMCast::blacklist(const gmcast::Proto* proto)
+{
+    initial_addrs_.erase(proto->remote_addr());
+    pending_addrs_.erase(proto->remote_addr());
+    addr_blacklist_.insert(std::make_pair(
+                               proto->remote_addr(),
+                               AddrEntry(gu::datetime::Date::now(),
+                                         gu::datetime::Date::now(),
+                                         proto->remote_uuid())));
+}
+
+bool gcomm::GMCast::is_not_own_and_duplicate_exists(
+    const Proto* proto) const
+{
+    assert(proto->remote_uuid() != gcomm::UUID::nil());
+    const Proto* other(find_other_local_endpoint(*proto_map_, proto));
+    if (!other)
+    {
+        // Not own
+        // Check if remote UUID matches to self
+        if (proto->remote_uuid() == uuid())
+        {
+            return true;
+        }
+        // Check if other proto entry with same remote
+        // UUID but different remote address exists.
+        other = find_other_endpoint_same_remote_uuid(*proto_map_, proto);
+        if (other && other->remote_addr() != proto->remote_addr())
+        {
+            return true;
+        }
+    }
+    return false;
 }
 
 // Erase proto entry in safe manner
@@ -389,7 +503,7 @@ void gcomm::GMCast::gmcast_accept()
         return;
     }
 
-    if (isolate_ == true)
+    if (is_isolated(isolate_))
     {
         log_debug << "dropping accepted socket due to isolation";
         tp->close();
@@ -551,6 +665,8 @@ void gcomm::GMCast::handle_established(Proto* est)
     log_info << self_string() << " connection established to "
              << est->remote_uuid() << " "
              << est->remote_addr();
+    // UUID checks are handled during protocol handshake
+    assert(est->remote_uuid() != uuid());
 
     if (is_evicted(est->remote_uuid()))
     {
@@ -560,34 +676,6 @@ void gcomm::GMCast::handle_established(Proto* est)
         return;
     }
 
-    if (est->remote_uuid() == uuid())
-    {
-        std::set<std::string>::iterator
-            ia_i(initial_addrs_.find(est->remote_addr()));
-        if (ia_i != initial_addrs_.end())
-        {
-            initial_addrs_.erase(ia_i);
-        }
-        AddrList::iterator i(pending_addrs_.find(est->remote_addr()));
-        if (i != pending_addrs_.end())
-        {
-            if (addr_blacklist_.find(est->remote_addr()) == addr_blacklist_.end())
-            {
-                log_warn << self_string()
-                         << " address '" << est->remote_addr()
-                         << "' points to own listening address, blacklisting";
-            }
-            pending_addrs_.erase(i);
-            addr_blacklist_.insert(make_pair(est->remote_addr(),
-                                             AddrEntry(gu::datetime::Date::now(),
-                                                       gu::datetime::Date::now(),
-                                                       est->remote_uuid())));
-        }
-        erase_proto(proto_map_->find_checked(est->socket()->id()));
-        update_addresses();
-        return;
-    }
-
     // If address is found from pending_addrs_, move it to remote_addrs list
     // and set retry cnt to -1
     const std::string& remote_addr(est->remote_addr());
@@ -734,7 +822,6 @@ void gcomm::GMCast::handle_failed(Proto* failed)
     update_addresses();
 }
 
-
 bool gcomm::GMCast::is_connected(const std::string& addr, const UUID& uuid) const
 {
     for (ProtoMap::const_iterator i = proto_map_->begin();
@@ -944,7 +1031,7 @@ void gcomm::GMCast::update_addresses()
 
 void gcomm::GMCast::reconnect()
 {
-    if (isolate_ == true)
+    if (is_isolated(isolate_))
     {
         log_debug << "skipping reconnect due to isolation";
         return;
@@ -1386,15 +1473,16 @@ void gcomm::GMCast::handle_up(const void*        id,
                     p->set_tstamp(gu::datetime::Date::now());
                     gu_trace(p->handle_message(msg));
                 }
-                catch (gu::Exception& e)
+                catch (const gu::Exception& e)
                 {
-                    log_warn << "handling gmcast protocol message failed: "
-                             << e.what();
                     handle_failed(p);
                     if (e.get_errno() == ENOTRECOVERABLE)
                     {
                         throw;
                     }
+                    log_warn
+                        << "handling gmcast protocol message failed: "
+                        << e.what();
                     return;
                 }
 
@@ -1403,7 +1491,7 @@ void gcomm::GMCast::handle_up(const void*        id,
                     handle_failed(p);
                     return;
                 }
-                else if (p->changed() == true)
+                else if (p->check_changed_and_reset() == true)
                 {
                     update_addresses();
                     check_liveness();
@@ -1579,6 +1667,7 @@ void gcomm::GMCast::handle_stable_view(const View& view)
             }
             i = i_next;
         }
+        prim_view_reached_ = true;
     }
     else if (view.type() == V_REG)
     {
@@ -1685,7 +1774,8 @@ void gcomm::GMCast::add_or_del_addr(const std::string& val)
 }
 
 
-bool gcomm::GMCast::set_param(const std::string& key, const std::string& val)
+bool gcomm::GMCast::set_param(const std::string& key, const std::string& val,
+                              Protolay::sync_param_cb_t& sync_param_cb)
 {
     try
     {
@@ -1712,10 +1802,18 @@ bool gcomm::GMCast::set_param(const std::string& key, const std::string& val)
         }
         else if (key == Conf::GMCastIsolate)
         {
-            isolate_ = gu::from_string<bool>(val);
+            int tmpval = gu::from_string<int>(val);
+            if (tmpval < 0 || tmpval > 2)
+            {
+                gu_throw_error(EINVAL)
+                    << "invalid value for gmacst.isolate: '"
+                    << tmpval << "'";
+            }
+            isolate_ = tmpval;
             log_info << "turning isolation "
-                     << (isolate_ == true ? "on" : "off");
-            if (isolate_ == true)
+                     << (isolate_ == 1 ? "on" :
+                         (isolate_ == 2 ? "force quit" : "off"));
+            if (isolate_)
             {
                 // delete all entries in proto map
                 ProtoMap::iterator pi, pi_next;
